from typing import Callable

import numpy as np
from scipy.signal import find_peaks as find_peaks_orig


def assert_same_shape(*arrays: np.ndarray) -> None:
    shapes = [x.shape for x in arrays]
    shape_set = set(shapes)
    assert (
        len(shape_set) == 1
    ), f"Expected all tensors to have the same shape, but shapes are: {shapes}"


def find_peaks(x: np.ndarray, distance: int, ignore_peaks_at_boundary: bool = False) -> np.ndarray:
    """
    Improved `find_peaks` including a work-around for the strange handling of neighborhoods:
    https://github.com/scipy/scipy/issues/18495

    https://docs.scipy.org/doc/scipy-1.12.0/reference/generated/scipy.signal.find_peaks.html#scipy.signal.find_peaks
    """

    peak_indices, _properties = find_peaks_orig(x, distance=distance)

    selected_peak_indices = []

    # To fix the semantics of `find_peaks` we manually check the dominance over the local
    # neighborhood.
    #
    # Note: The reason not to use `argrelmax` was that it does not detect peaks if a peak repeats
    # the same value multiple times. Apparently such peaks are not at all considered peaks by
    # `argrelmax`. In my use cases that would be fatal, because there could be strong "peaks",
    # which happen to repeat the same value just due to noise. They definitely should be detected
    # anyway. See reference:
    # https://docs.scipy.org/doc/scipy-1.12.0/reference/generated/scipy.signal.argrelmax.html

    for peak_idx in peak_indices:
        if ignore_peaks_at_boundary:
            # Ignore peaks if they are exactly at the boundary
            if peak_idx == 0 or peak_idx == len(x) - 1:
                continue

        neighborhood_left = np.s_[max(peak_idx - distance, 0) : peak_idx]
        neighborhood_right = np.s_[peak_idx + 1 : peak_idx + distance + 1]
        neighborhood_max_left = np.max(x[neighborhood_left])
        neighborhood_max_right = np.max(x[neighborhood_right])

        # If a peak does not satisfy the local dominance criterion we don't even consider
        # it a peak.
        if x[peak_idx] < neighborhood_max_left or x[peak_idx] < neighborhood_max_right:
            continue

        selected_peak_indices.append(peak_idx)

    return np.array(selected_peak_indices)


def expspace(value_from: float, value_upto: float, n: int, grow_factor: float) -> np.ndarray:
    """
    Function similar to np.linspace / np.logspace, but with a slightly more convenient interface
    than logspace: It allows to specify a `grow_factor` which determines how much larger/smaller
    each interval is to the next one.

    For instance, if the `grow_factor` is 0.5, the second interval is half the size of the first,
    the third half the size of the second, etc.

    For lack of a better word, called 'expspace'...
    """
    if n <= 1:
        raise ValueError(f"Number of points must be >= 2, got: {n}")

    scaled = np.cumsum(np.cumprod(np.full(n, grow_factor)))

    scaled_min = np.min(scaled)
    scaled_max = np.max(scaled)
    delta_scaled = scaled_max - scaled_min
    delta_target = value_upto - value_from
    return (scaled - scaled_min) / delta_scaled * delta_target + value_from


def expspace_delta(value_from: float, value_upto: float, n: int, delta: float) -> np.ndarray:
    """
    Similar to the function above, but allows to specify the size of the first interval (`delta`)
    instead of the growth factor.

    The idea here is to solve this equation for the growth factor `f` given a delta `d`, and a
    range `r`:

        n - 1
         ___
         ╲         i
         ╱    d ⋅ f  = r
         ‾‾‾
        i = 0

    We can pull out the `d`

            n - 1
             ___
             ╲     i
        d ⋅  ╱    f  = r
             ‾‾‾
            i = 0

    bring it to the other side, and write it using `c`.

            r
        c = ─
            d

    The sum then can be rewritten using the formula for the geometric series.

             N
        1 - f
        ────── = c
         1 - f

    We now have to solve for `f`, which unfortunately involves a higher-order polynomial.
    According to the following this may be solvable analytically, but a numeric solution
    should be good enough for now:

    https://math.stackexchange.com/questions/1437161/how-to-solve-nth-degree-polynomial-equation-with-terms-only-in-n-1-and-0

    (Math written with: https://arthursonzogni.com/Diagon/#Math)
    """

    value_range = value_upto - value_from

    if n <= 1:
        raise ValueError(f"Number of points must be >= 2, got: {n}")
    elif n == 2:
        if delta == value_range:
            return np.array([value_from, value_upto])
        else:
            raise ValueError("For n = 1, the delta must match the value range")

    if delta == 0.0:
        raise ValueError("Delta must not be zero.")

    if value_range > 0.0:
        if delta < 0.0:
            raise ValueError("Delta must not be negative, if range is positive.")
    elif value_range < 0.0:
        if delta > 0.0:
            raise ValueError("Delta must not be positive, if range is negative.")
    else:
        raise ValueError("Value range must not be zero.")

    if abs(delta) >= abs(value_range):
        raise ValueError("Delta must not be larger than value range.")

    c = value_range / delta
    assert c > 1.0

    # In terms of the geometric series we care about the number of terms/interval,
    # not the number of points to generate.
    N = n - 1

    def f(x: float) -> float | None:
        # Return the value of geometric series up to term N.
        if x == 1.0:
            return N
        else:
            try:
                return (1 - x**N) / (1 - x)
            except OverflowError:
                return None

    ratio = delta / (value_range / N)
    if ratio < 1.0:
        # Growing case
        x_l = 1.0
        x_r = c
        # Evaluating the geometric series for large factors can overflow in the `x**N`
        # term. We bring the factor down closer to 1.0 via square-rooting until we have
        # a valid value.
        while f(x_r) is None:
            x_r = np.sqrt(x_r)
    elif ratio > 1.0:
        # Shrinking case
        x_l = 0.0
        x_r = 1.0
    else:
        return expspace(value_from, value_upto, n, grow_factor=1.0)

    f_l = f(x_l)
    f_r = f(x_r)
    assert f_l is not None
    assert f_r is not None
    assert f_l - c < 0.0, f"Expected `f(x_l) - c` to be negative, but is {f_l - c} ({x_l=}, {c=})"
    assert f_r - c > 0.0, f"Expected `f(x_r) - c` to be positive, but is {f_r - c} ({x_r=}, {c=})"

    # Initially I was using Newton-Raphson instead, but since the equation has two
    # roots, there is a danger of finding the wrong solution. Bisecting can avoid
    # that be initializing properly.
    x = bisect(lambda x: f(x), c, x_l, x_r)

    return expspace(value_from, value_upto, n, grow_factor=x)


Func = Callable[[float], float | None]


def bisect(
    f: Func, target: float, x_l: float, x_r: float, tol: float = 1e-8, max_iter: int = 10000
) -> float:

    for _ in range(max_iter):
        x_mid = (x_l + x_r) * 0.5
        f_x_mid = f(x_mid)

        if f_x_mid is None:
            raise RuntimeError(f"f has overflowed for {x_mid=}")

        delta = f_x_mid - target

        if abs(delta) < tol:
            return x_mid

        if delta < 0.0:
            x_l = x_mid
        else:
            x_r = x_mid

    raise ValueError("Maximum iterations exceeded. No solution found.")



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Generic, TypeVar

T = TypeVar("T")


@dataclass
class Lazy(Generic[T]):

    def __init__(self, factory: Callable[[], T]):
        self._factory: Callable[[], T] = factory
        self._value: T | None = None

    def __call__(self) -> T:
        if self._value is None:
            self._value = self._factory()
        return self._value

    @staticmethod
    def get(x: T | Lazy[T]) -> T:
        if isinstance(x, Lazy):
            return x()
        else:
            return x



from contextlib import contextmanager
from typing import Generator, Iterator, Sequence, TypeVar

import torch
from torch.testing._comparison import assert_close as assert_close_orig
from torch.utils.data import DataLoader

T = TypeVar("T")

# -----------------------------------------------------------------------------
# For general type safety
# -----------------------------------------------------------------------------


class TypeSafeModule(torch.nn.Module):
    # def __getattr__(self, name: str) -> torch.Tensor | torch.nn.Module:
    def __getattr__(self, name: str) -> object:
        # Type-safety work around for https://github.com/pytorch/pytorch/pull/104321
        return super().__getattr__(name)


# -----------------------------------------------------------------------------
# Type-safe data loading
# -----------------------------------------------------------------------------


def iter_dataloader(data_loader: DataLoader[T]) -> Iterator[T]:
    """
    Work-around for: https://github.com/pytorch/pytorch/issues/119123
    """
    for batch in data_loader:
        yield batch


# -----------------------------------------------------------------------------
# Misc utils
# -----------------------------------------------------------------------------


@contextmanager
def local_eval(module: torch.nn.Module) -> Generator[None, None, None]:
    """
    Allows to switch module to `eval` mode with automatic restoring of original state.
    """
    training_state_before = module.training
    module.eval()
    try:
        yield
    finally:
        module.train(training_state_before)


def inverse_sigmoid(x: torch.Tensor) -> torch.Tensor:
    # https://stackoverflow.com/a/77031136/1804173
    return torch.log(x) - torch.log(1 - x)


# -----------------------------------------------------------------------------
# Testing utils
# -----------------------------------------------------------------------------


def assert_close(
    actual: torch.Tensor,
    expected: torch.Tensor | Sequence[object] | float | int,
    rtol: float | None = None,
    atol: float | None = None,
    check_dtype: bool = True,
) -> None:
    if not isinstance(expected, torch.Tensor):
        expected = torch.tensor(expected)
    print(f"\nActual:   {actual}")
    print(f"Expected: {expected}")
    assert_close_orig(actual, expected, rtol=rtol, atol=atol, check_dtype=check_dtype)



from typing import NamedTuple

import torch
from torch.utils.data import DataLoader, Dataset
from typing_extensions import assert_type

from pzen import torch_utils
from pzen.torch_utils import assert_close


def test_iter_dataloader():

    class Batch(NamedTuple):
        x: torch.Tensor

    class MyDataset(Dataset):
        def __len__(self) -> int:
            return 3

        def __getitem__(self, idx: int) -> Batch:
            return Batch(x=torch.tensor(idx))

    data_loader: DataLoader[Batch] = DataLoader(MyDataset())

    for batch in torch_utils.iter_dataloader(data_loader):
        assert_type(batch, Batch)
        assert isinstance(batch, Batch)


def test_inverse_sigmoid__basic():
    assert_close(torch_utils.inverse_sigmoid(torch.tensor(0.9)), +2.1972246170043945)
    assert_close(torch_utils.inverse_sigmoid(torch.tensor(0.5)), 0.0)
    assert_close(torch_utils.inverse_sigmoid(torch.tensor(0.1)), -2.1972246170043945)


def test_inverse_sigmoid__shape_preserving():
    x = torch.tensor(
        [
            [-1.0, 0.0, -7.0, 2.0],
            [3.0, 2.0, 4.0, -5.0],
        ]
    )
    sigmoid_x = torch.sigmoid(x)
    print(sigmoid_x)
    assert_close(torch_utils.inverse_sigmoid(sigmoid_x), x)



from pzen.lazy_utils import Lazy


def test_lazy():

    num_calls = 0

    def costly_func() -> float:
        nonlocal num_calls
        num_calls += 1
        return 123.456

    def example_usage(
        # Usage should in general allow both eager and lazy passing, i.e., `T | Lazy[T]`,
        # and the default value is expressed as a Lazy instance to avoid its evaluation at
        # function declaration time, different from just `arg: float = costly_func()`.
        arg: float | Lazy[float] = Lazy(lambda: costly_func()),
    ):
        # Using `Lazy.get` makes it convenient to work with `T | Lazy[T]`.
        assert Lazy.get(arg) == 123.456

    assert num_calls == 0

    example_usage()
    assert num_calls == 1

    example_usage()
    assert num_calls == 1



import numpy as np

from pzen.cache_utils import _CACHE_DIR, with_memory

_CALLS: list[tuple[float, str, np.ndarray]] = []


@with_memory
def f(a: float, b: str, c: np.ndarray) -> float:
    _CALLS.append((a, b, c))
    return (a + len(b)) * np.sum(c)


def test_cache_utils():
    f.clear()  # type: ignore[attr-defined] # MemorizedFunc magic to clear the cache.

    assert f(1.0, "foo", np.array([1, 2, 3])) == 24.0
    assert len(_CALLS) == 1
    assert f(1.0, "foo", np.array([1, 2, 3])) == 24.0
    assert len(_CALLS) == 1

    assert f(2.0, "foo", np.array([1, 2, 3])) == 30.0
    assert len(_CALLS) == 2
    assert f(2.0, "foo", np.array([1, 2, 3])) == 30.0
    assert len(_CALLS) == 2

    assert f(2.0, "fooo", np.array([1, 2, 3])) == 36.0
    assert len(_CALLS) == 3
    assert f(2.0, "fooo", np.array([1, 2, 3])) == 36.0
    assert len(_CALLS) == 3

    assert f(1.0, "foo", np.array([1, 2, 4])) == 28.0
    assert len(_CALLS) == 4
    assert f(1.0, "foo", np.array([1, 2, 4])) == 28.0
    assert len(_CALLS) == 4

    assert _CACHE_DIR.exists()



from pathlib import Path

from matplotlib.axes import Axes
from matplotlib.figure import Figure
from typing_extensions import assert_type

from pzen import mpl_utils


def test_subplots_wrapper__plot(tmp_path: Path):
    plot_path = tmp_path / "subdir" / "plot.png"
    with mpl_utils.plot(save_as=plot_path) as (fig, ax):
        assert_type(fig, Figure)
        assert isinstance(fig, Figure)
        assert_type(ax, Axes)
        assert isinstance(ax, Axes)

        ax.plot([1, 2, 3])

    assert plot_path.is_file()


def test_subplots_wrapper__plot_rows(tmp_path: Path):
    plot_path = tmp_path / "subdir" / "plot.png"
    with mpl_utils.plot_rows(3, save_as=plot_path) as (fig, axes):
        assert_type(fig, Figure)
        assert isinstance(fig, Figure)
        assert_type(axes, list[Axes])

        axes[0].plot([1, 2, 3])
        axes[1].plot([1, 2, 3])
        axes[2].plot([1, 2, 3])

    assert plot_path.is_file()


def test_subplots_wrapper__plot_cols(tmp_path: Path):
    plot_path = tmp_path / "subdir" / "plot.png"
    with mpl_utils.plot_cols(3, save_as=plot_path) as (fig, axes):
        assert_type(fig, Figure)
        assert isinstance(fig, Figure)
        assert_type(axes, list[Axes])

        axes[0].plot([1, 2, 3])
        axes[1].plot([1, 2, 3])
        axes[2].plot([1, 2, 3])

    assert plot_path.is_file()



import datetime
import tempfile
from pathlib import Path
from typing import Any, Callable, TypeVar

from joblib import Memory
from typing_extensions import ParamSpec
from yachalk import chalk

_CACHE_DIR = Path(tempfile.gettempdir()) / "joblib_cache"
_MEMORY = Memory(
    _CACHE_DIR,
)


P = ParamSpec("P")
R = TypeVar("R")


def with_memory(f: Callable[P, R]) -> Callable[P, R]:
    """
    Decorator to apply joblib's Memory in a type-safe manner:
    https://joblib.readthedocs.io/en/latest/memory.html
    """

    def cache_validation_cb(metadata: dict[str, Any]) -> bool:
        time = datetime.datetime.fromtimestamp(metadata["time"]).strftime("%Y-%m-%d %H:%M:%S")
        duration = metadata["duration"]
        print(
            chalk.bold(
                f"Using cached result for `{f.__name__}` (computed at {time}, in {duration:.3f} s)"
            )
        )
        return True

    return _MEMORY.cache(
        verbose=0,
        cache_validation_callback=cache_validation_cb,
    )(
        f  # pyright: ignore
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import numpy as np
from typing_extensions import assert_never

# -----------------------------------------------------------------------------
# Low-level API
# -----------------------------------------------------------------------------


def normalize(x: np.ndarray) -> np.ndarray:
    """
    Normalize signal to stay within [-1, +1], but leave the zero point unmodified.
    """
    abs_max = np.abs(x).max()
    if abs_max != 0:
        return x / abs_max
    else:
        return x


def normalize_min_max(x: np.ndarray, min: float = -1.0, max: float = 1.0) -> np.ndarray:
    """
    This function normalizes the min/max to [-1, +1].

    Note that this can be a bit confusing, because it may look like the output has a constant
    DC-offset.
    """
    orig_max = x.max()
    orig_min = x.min()

    if orig_max > orig_min:
        return min + (x - orig_min) / (orig_max - orig_min) * (max - min)
    else:
        return x


def pad_to_multiple_of(x: np.ndarray, block_size: int) -> np.ndarray:
    """
    Right-pad signal so that its length is a multiple of `block_size`.
    """
    assert block_size > 0
    n_pad = (block_size - len(x) % block_size) % block_size
    x = np.pad(x, (0, n_pad))
    return x


# -----------------------------------------------------------------------------
# High-level API
# -----------------------------------------------------------------------------


@dataclass
class Samples:
    value: int


@dataclass
class Seconds:
    value: float


"""
Note on naming: Alternative names considered 'Duration' or 'Length'. While most usages
actually are semantically more a "duration" or "length" there can also be usages that
are semantically an "offset", which would make duration/length awkward. `t: Time` is
short and works reasonable well for both.
"""
Time = Samples | Seconds


def in_samples(t: Time | None, sr: int) -> int:
    if t is None:
        return 0
    elif isinstance(t, Samples):
        return t.value
    elif isinstance(t, Seconds):
        return round(t.value * sr)
    else:
        assert_never(t)


LinLog = Literal["lin", "log"]
LinExp = Literal["lin", "exp"]

Quantity = Literal["root-power", "power"]


@dataclass
class Signal:
    """
    High-level wrapper of signals.

    Note on naming: Alternative names considered 'Audio' or 'Wave'. However, this class
    is more general than audio/wave signals, because it could also be used for other
    kind of signals like amplitude envelops, frequency sequences, phases, etc.
    """

    x: np.ndarray
    sr: int

    def __post_init__(self) -> None:
        # Invariants
        assert self.x.ndim == 1, f"Signals must have a dimension of 1, but is {self.x.ndim}"
        assert self.sr > 0

    # Construction helpers

    @staticmethod
    def empty(sr: int) -> Signal:
        return Signal(x=np.array([]), sr=sr)

    @staticmethod
    def zeros(t: Time, sr: int) -> Signal:
        n = in_samples(t, sr)
        return Signal(x=np.zeros(n), sr=sr)

    @staticmethod
    def ones(t: Time, sr: int) -> Signal:
        n = in_samples(t, sr)
        return Signal(x=np.ones(n), sr=sr)

    @staticmethod
    def full(t: Time, value: float, sr: int) -> Signal:
        n = in_samples(t, sr)
        return Signal(x=np.full(n, value), sr=sr)

    # Basic operators

    def __len__(self) -> int:
        return len(self.x)

    def __add__(self, other: Signal) -> Signal:
        assert (
            self.sr == other.sr
        ), f"Can only add signals of same sample rate, but {self.sr} != {other.sr}"
        assert len(self) == len(other), (
            f"Trying to add signals with different lengths ({len(self)} != {len(other)}). "
            "Use `mix_at` if implicit length extension is desired."
        )
        return Signal(x=self.x + other.x, sr=self.sr)

    def __mul__(self, other: Signal) -> Signal:
        assert (
            self.sr == other.sr
        ), f"Can only multiply signals of same sample rate, but {self.sr} != {other.sr}"
        assert len(self) == len(
            other
        ), f"Trying to multiply signals with different lengths ({len(self)} != {len(other)})."
        return Signal(x=self.x * other.x, sr=self.sr)

    # Misc

    @property
    def unpack(self) -> tuple[np.ndarray, int]:
        return self.x, self.sr

    def in_samples(self, t: Time | None) -> int:
        return in_samples(t, self.sr)

    def len(self) -> int:
        """Convenience alternative to __len__ for more fluent API."""
        return len(self)

    def len_in_sec(self) -> float:
        return len(self) / self.sr

    # Basic element-wise (shape preserving) operations

    def scale(self, factor: float) -> Signal:
        return Signal(factor * self.x, self.sr)

    def abs(self) -> Signal:
        return Signal(np.abs(self.x), self.sr)

    def square(self) -> Signal:
        return Signal(self.x * self.x, self.sr)

    def normalize(self) -> Signal:
        return Signal(normalize(self.x), self.sr)

    def normalize_min_max(self, min: float = -1.0, max: float = 1.0) -> Signal:
        return Signal(normalize_min_max(self.x, min, max), self.sr)

    def into_exp_envelope(self, min_db: float = -80.0, quantity: Quantity = "root-power") -> Signal:
        """
        Converts a given (linear) envelope, assumed to be in [0.0, 1.0] to an equivalent _exponentially_
        scaled envelope to counter the _logarithmic_ perception of loudness.

        Due to the nature of the conversion, a minimum level has to be specified, which is easiest
        to do in terms of decibel. Downside: The conversion from decibel to a plain factor requires
        to clarify to quantity semantics, i.e., "root-power quantity" or "power quantity". Since
        envelopes typically operate on amplitude level, which are root-power quantities, we default
        to "root-power".
        """
        decibels = self.normalize_min_max(min_db, 0.0).x
        if quantity == "root-power":
            factors = 10 ** (decibels / 20)
        elif quantity == "power":
            factors = 10 ** (decibels / 10)
        else:
            assert_never(quantity)
        return Signal(x=factors, sr=self.sr)

    # Length affecting operations

    def pad(self, *, pad_l: Time | None = None, pad_r: Time | None = None) -> Signal:
        n_pad_l = self.in_samples(pad_l)
        n_pad_r = self.in_samples(pad_r)
        x = np.pad(self.x, (n_pad_l, n_pad_r))
        return Signal(x, self.sr)

    def pad_to_multiple_of(self, block_size: Time) -> Signal:
        x = pad_to_multiple_of(self.x, block_size=self.in_samples(block_size))
        return Signal(x, self.sr)

    def concat(self, other: Signal) -> Signal:
        assert (
            self.sr == other.sr
        ), f"Can only concatenate signals of same sample rate, but {self.sr} != {other.sr}"
        return Signal(x=np.concatenate([self.x, other.x]), sr=self.sr)

    # Advanced operations

    def mix_at(self, offset: Time, other: Signal, allow_extend: bool = True) -> Signal:
        assert (
            self.sr == other.sr
        ), f"Can only mix signals of same sample rate, but {self.sr} != {other.sr}"

        offset_index = self.in_samples(offset)
        if offset_index < 0:
            raise ValueError(f"Negative offset indices ({offset_index}) are not supported")
        required_len = offset_index + other.len()

        if required_len > self.len():
            if not allow_extend:
                raise ValueError(
                    f"Mixing signal would require a length of {required_len}, "
                    f"but signal only has length {self.len()}"
                )
            x = np.pad(self.x, (0, required_len - self.len()))
        else:
            x = self.x.copy()

        x[offset_index : offset_index + other.len()] += other.x
        return Signal(x=x, sr=self.sr)

    # Envelope modulations

    def envelope_ramped(self, kind: LinExp, t_l: Time, t_r: Time | None = None) -> Signal:
        gen = SignalGenerator(self.sr)
        envelope = gen.envelope_ramped(Samples(self.len()), t_l, t_r)
        if kind == "exp":
            envelope = envelope.into_exp_envelope()
        return envelope * self


# -----------------------------------------------------------------------------
# Generator API
# -----------------------------------------------------------------------------


DEFAULT_TIME = Seconds(1.0)


class SignalGenerator:
    def __init__(self, sr: int):
        self.sr = sr

    def _in_samples(self, t: Time) -> int:
        return in_samples(t, self.sr)

    def _make_signal(self, x: np.ndarray) -> Signal:
        return Signal(x=x, sr=self.sr)

    # Forwarded construction helpers

    def empty(self) -> Signal:
        return Signal.empty(sr=self.sr)

    def zeros(self, t: Time) -> Signal:
        return Signal.zeros(t, sr=self.sr)

    def ones(self, t: Time) -> Signal:
        return Signal.ones(t, sr=self.sr)

    def full(self, t: Time, value: float) -> Signal:
        return Signal.full(t, value, sr=self.sr)

    # General purpose generators

    def ramp(self, t: Time, start: float = 0.0, end: float = 1.0) -> Signal:
        n = self._in_samples(t)
        return self._make_signal(np.linspace(start, end, n))

    # Possible extension would be to have `multi_ramp` that takes a vararg of
    # tuples of (delta: Time, value: float), allowing to chain linear ramps

    # Envelope generators

    def envelope_ramped(self, t: Time, t_l: Time, t_r: Time | None = None) -> Signal:
        """
        Returns an envelope that ramps up from 0 -> 1 and back from 1 -> 0.
        """
        if t_r is None:
            t_r = t_l
        n = self._in_samples(t)
        n_l = self._in_samples(t_l)
        n_r = self._in_samples(t_r)
        ramp_l = np.linspace(0.0, 1.0, n_l + 1)[1:-1]
        ramp_r = np.linspace(1.0, 0.0, n_r + 1)[1:-1]

        n_l = len(ramp_l)
        n_r = len(ramp_r)

        if n_l + n_r <= n:
            x = np.ones(n)
            x[: len(ramp_l)] = ramp_l
            x[n - len(ramp_r) :] = ramp_r
            return self._make_signal(x)
        else:
            if n_l < n:
                x1 = np.pad(ramp_l, (0, n - n_l), constant_values=1.0)
            else:
                x1 = ramp_l[:n]
            if n_r < n:
                x2 = np.pad(ramp_r, (n - n_r, 0), constant_values=1.0)
            else:
                x2 = ramp_r[-n:]
            return self._make_signal(np.minimum(x1, x2))

    # Frequency generators

    def vibrato(
        self,
        t: Time,
        f: float,
        f_vibrato: float = 6.0,
        semitones: float = 0.3,
        t_l: Time = Seconds(0.1),
        t_r: Time | None = None,
    ) -> Signal:
        if t_r is None:
            t_r = t_l

        modulation_strength = Signal.ones(t, sr=self.sr).envelope_ramped("lin", t_l, t_r).x
        modulation = modulation_strength * self.sine(freq=f_vibrato, t=t).scale(semitones).x

        factors = 2 ** (modulation / 12.0)
        freqs = f * factors

        return self._make_signal(freqs)

    def sweep(
        self,
        t: Time,
        f1: float,
        f2: float,
    ) -> Signal:
        n = self._in_samples(t)
        ratio = f2 / f1
        freqs = f1 * np.exp(np.linspace(0.0, np.log(ratio), n))
        return self._make_signal(freqs)

    # Audio-like generators

    def silence(self, t: Time = DEFAULT_TIME) -> Signal:
        # Technically redundant to `zeros` but alias can express intent better.
        return Signal.zeros(t, self.sr)

    def sine(
        self, freq: float = 440.0, t: Time = DEFAULT_TIME, phase_offset: float = 0.0
    ) -> Signal:
        """
        Generate perfect sine from a frequency.
        """
        n = self._in_samples(t)
        ts = np.arange(n) / self.sr
        x = np.sin(phase_offset + 2.0 * np.pi * freq * ts)
        return self._make_signal(x)

    def sine_from_freqs(self, freqs: Signal, phase_offset: float = 0.0) -> Signal:
        assert self.sr == freqs.sr
        phases = np.cumsum(freqs.x) / self.sr
        x = np.sin(phase_offset + 2.0 * np.pi * phases)
        return self._make_signal(x)



import gc
from collections.abc import Iterable
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from .core_types import StrPath

_DEFAULT_FIGSIZE = (12, 8)


@contextmanager
def plot(
    figsize: tuple[float, float] = _DEFAULT_FIGSIZE,
    fig_title: str | None = None,
    interactive: bool = False,
    save_as: StrPath | None = None,
    run_gc_collection: bool = False,
) -> Generator[tuple[Figure, Axes], None, None]:
    with _subplots_wrapper(
        nrows=1,
        ncols=1,
        figsize=figsize,
        fig_title=fig_title,
        interactive=interactive,
        save_as=save_as,
        sharex=False,
        sharey=False,
        run_gc_collection=run_gc_collection,
    ) as (fig, axes):
        assert isinstance(axes, Axes)
        yield fig, axes


@contextmanager
def plot_rows(
    nrows: int,
    figsize: tuple[float, float] = _DEFAULT_FIGSIZE,
    fig_title: str | None = None,
    interactive: bool = False,
    save_as: StrPath | None = None,
    sharex: bool = True,
    run_gc_collection: bool = False,
) -> Generator[tuple[Figure, list[Axes]], None, None]:
    with _subplots_wrapper(
        nrows=nrows,
        ncols=1,
        figsize=figsize,
        fig_title=fig_title,
        interactive=interactive,
        save_as=save_as,
        sharex=sharex,
        sharey=False,
        run_gc_collection=run_gc_collection,
    ) as (fig, axes):
        # Note that matplotlib returns a numpy array, which is type erasing and
        # has no practical benefits to me. Therefore, convert to a list.
        axes = list(axes)
        assert all(isinstance(ax, Axes) for ax in axes)
        yield fig, axes


@contextmanager
def plot_cols(
    ncols: int,
    figsize: tuple[float, float] = _DEFAULT_FIGSIZE,
    fig_title: str | None = None,
    interactive: bool = False,
    save_as: StrPath | None = None,
    sharey: bool = True,
    run_gc_collection: bool = False,
) -> Generator[tuple[Figure, list[Axes]], None, None]:
    with _subplots_wrapper(
        nrows=1,
        ncols=ncols,
        figsize=figsize,
        fig_title=fig_title,
        interactive=interactive,
        save_as=save_as,
        sharex=False,
        sharey=sharey,
        run_gc_collection=run_gc_collection,
    ) as (fig, axes):
        # Note that matplotlib returns a numpy array, which is type erasing and
        # has no practical benefits to me. Therefore, convert to a list.
        axes = list(axes)
        assert all(isinstance(ax, Axes) for ax in axes)
        yield fig, axes


@contextmanager
def _subplots_wrapper(
    nrows: int,
    ncols: int,
    figsize: tuple[float, float],
    fig_title: str | None,
    interactive: bool,
    save_as: StrPath | None,
    sharex: bool,
    sharey: bool,
    run_gc_collection: bool,
) -> Generator[tuple[Figure, Any], None, None]:

    fig, axes = plt.subplots(
        nrows,
        ncols,
        figsize=figsize,
        sharex=sharex,
        sharey=sharey,
    )
    if fig_title is not None:
        fig.suptitle(fig_title)

    if isinstance(axes, Iterable):
        for ax in axes:
            ax.grid()
    else:
        axes.grid()

    try:
        yield fig, axes
    finally:
        fig.tight_layout()

        if interactive:
            plt.show()

        if save_as is not None:
            Path(save_as).parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(str(save_as))

        plt.close(fig)

        # Work-around for plotting in a multi-processing environment (e.g. during
        # using a torch data loader).
        if run_gc_collection:
            gc.collect()



import numpy as np
import numpy.testing as npt
import pytest

from pzen.numpy_utils import expspace, expspace_delta, find_peaks


def test_find_peaks__basics():
    npt.assert_equal(
        find_peaks(np.array([1.0, 2.0, 0.2, 3.0, 2.3]), distance=1),
        np.array([1, 3]),
    )
    npt.assert_equal(
        find_peaks(np.array([1.0, 2.0, 0.2, 3.0, 2.3]), distance=2),
        np.array([3]),
    )
    npt.assert_equal(
        find_peaks(np.array([1.0, 2.0, 0.2, 3.0, 2.3]), distance=10),
        np.array([3]),
    )


def test_find_peaks__repeated_peak_value():
    npt.assert_equal(
        find_peaks(np.array([1.0, 1.0, 3.0, 3.0, 1.0, 1.0]), distance=1),
        np.array([2]),
    )


def test_find_peaks__test_case_from_issue_18495():
    # https://github.com/scipy/scipy/issues/18495
    x = np.zeros(200)
    x[100:] = np.linspace(1.0, 0.0, 100) + np.random.normal(0.0, 1e-2, size=100)

    assert len(find_peaks(x, distance=10)) == 1


def test_expspace():
    npt.assert_allclose(
        expspace(0, 1, n=2, grow_factor=2.0),
        np.array([0.0, 1.0]),
    )
    npt.assert_allclose(
        expspace(5, 8, n=2, grow_factor=2.0),
        np.array([5.0, 8.0]),
    )
    npt.assert_allclose(
        expspace(-5, -8, n=2, grow_factor=2.0),
        np.array([-5.0, -8.0]),
    )

    npt.assert_allclose(
        expspace(0, 1, n=3, grow_factor=2.0),
        np.array([0.0, 1 / 3, 1.0]),
    )
    npt.assert_allclose(
        expspace(0, 1, n=3, grow_factor=0.5),
        np.array([0.0, 2 / 3, 1.0]),
    )

    npt.assert_allclose(
        expspace(5, 8, n=3, grow_factor=2.0),
        np.array([5.0, 6.0, 8.0]),
    )
    npt.assert_allclose(
        expspace(5, 8, n=3, grow_factor=0.5),
        np.array([5.0, 7.0, 8.0]),
    )

    npt.assert_allclose(
        expspace(-5, -8, n=3, grow_factor=2.0),
        np.array([-5.0, -6.0, -8.0]),
    )
    npt.assert_allclose(
        expspace(-5, -8, n=3, grow_factor=0.5),
        np.array([-5.0, -7.0, -8.0]),
    )


def test_expspace_delta__n_2():
    npt.assert_allclose(
        expspace_delta(5.0, 6.0, n=2, delta=1.0),
        np.array([5.0, 6.0]),
    )
    with pytest.raises(ValueError):
        expspace_delta(5.0, 6.0, n=2, delta=2.0)


def test_expspace_delta__positive_growing():
    npt.assert_allclose(
        expspace_delta(5.0, 8.0, n=3, delta=1.0),
        np.array([5.0, 6.0, 8.0]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 6.5, n=3, delta=0.5),
        np.array([5.0, 5.5, 6.5]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 12.0, n=4, delta=1.0),
        np.array([5.0, 6.0, 8.0, 12.0]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 8.5, n=4, delta=0.5),
        np.array([5.0, 5.5, 6.5, 8.5]),
    )


def test_expspace_delta__positive_shrinking():
    npt.assert_allclose(
        expspace_delta(5.0, 6.5, n=3, delta=1.0),
        np.array([5.0, 6.0, 6.5]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 5.75, n=3, delta=0.5),
        np.array([5.0, 5.5, 5.75]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 6.75, n=4, delta=1.0),
        np.array([5.0, 6.0, 6.5, 6.75]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 5.875, n=4, delta=0.5),
        np.array([5.0, 5.5, 5.75, 5.875]),
    )


def test_expspace_delta__negative_growing():
    npt.assert_allclose(
        expspace_delta(5.0, 2.0, n=3, delta=-1.0),
        np.array([5.0, 4.0, 2.0]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 3.5, n=3, delta=-0.5),
        np.array([5.0, 4.5, 3.5]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, -2.0, n=4, delta=-1.0),
        np.array([5.0, 4.0, 2.0, -2.0]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 1.5, n=4, delta=-0.5),
        np.array([5.0, 4.5, 3.5, 1.5]),
    )


def test_expspace_delta__negative_shrinking():
    npt.assert_allclose(
        expspace_delta(5.0, 3.5, n=3, delta=-1.0),
        np.array([5.0, 4.0, 3.5]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 4.25, n=3, delta=-0.5),
        np.array([5.0, 4.5, 4.25]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 3.25, n=4, delta=-1.0),
        np.array([5.0, 4.0, 3.5, 3.25]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 4.125, n=4, delta=-0.5),
        np.array([5.0, 4.5, 4.25, 4.125]),
    )


def test_expspace_delta__neutral():
    npt.assert_allclose(
        expspace_delta(5.0, 7.0, n=3, delta=1.0),
        np.array([5.0, 6.0, 7.0]),
    )
    npt.assert_allclose(
        expspace_delta(5.0, 3.0, n=3, delta=-1.0),
        np.array([5.0, 4.0, 3.0]),
    )


@pytest.mark.parametrize("n", [3, 10, 100, 1000, 10000])
def test_expspace_delta__misc_1(n: int):
    result = expspace_delta(0.0, 1000.0, n=n, delta=10.0)
    npt.assert_allclose(
        result[:2],
        np.array([0.0, 10.0]),
    )
    npt.assert_allclose(
        result[-1],
        np.array(1000.0),
    )


def test_expspace_delta__misc_2():
    result = expspace_delta(0.0, 22050 * 4, n=100, delta=10.0)
    npt.assert_allclose(
        result[:2],
        np.array([0.0, 10.0]),
    )
    npt.assert_allclose(
        result[-1],
        np.array(22050 * 4),
    )



from pathlib import Path
from typing import Literal, overload

import librosa
import numpy as np
import soundfile
import torch

from .cache_utils import with_memory
from .core_types import StrPath
from .signal_utils import Signal, normalize, normalize_min_max


@with_memory
def load_resampled(
    path: StrPath,
    sr: int,
    offset_sec: float = 0.0,
    duration_sec: float | None = None,
    duration_samples: int | None = None,
) -> Signal:
    """
    Wrapper around librosa.load (because resampling can be somewhat slow, and I prefer
    the more explicit caching approach).
    """
    print(f"Loading audio from '{path}' at sample rate {sr}...")
    signal, sampling_rate_loaded = librosa.load(
        path, sr=sr, offset=offset_sec, duration=duration_sec
    )
    assert sampling_rate_loaded == sr
    if duration_samples is not None:
        signal = signal[:duration_samples]
    return Signal(x=signal, sr=sr)


@overload
def soundfile_write(
    path: StrPath,
    data: np.ndarray | torch.Tensor,
    sr: int,
    *,
    clip: Literal["normalize", "normalize_min_max", "ignore", "warn", "error"] = "warn",
) -> None: ...


@overload
def soundfile_write(
    path: StrPath,
    data: Signal,
    *,
    clip: Literal["normalize", "normalize_min_max", "ignore", "warn", "error"] = "warn",
) -> None: ...


def soundfile_write(
    path: StrPath,
    data: np.ndarray | torch.Tensor | Signal,
    sr: int | None = None,
    *,
    clip: Literal["normalize", "normalize_min_max", "ignore", "warn", "error"] = "warn",
) -> None:
    """
    Wrapper for soundfile.write with improvements:
    - allow normalization (apparently default behavior is to clip, see
      https://stackoverflow.com/q/69388531/1804173)
    - warn if clip
    - nan check
    - torch support
    - support pathlib and str
    - output path creation
    """

    if isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()
    elif isinstance(data, Signal):
        sr = data.sr
        data = data.x

    assert sr is not None
    assert isinstance(data, np.ndarray)
    assert np.all(np.isfinite(data)), "Audio data contains NaNs or Infs."

    if clip == "normalize":
        data = normalize(data)

    elif clip == "normalize_min_max":
        data = normalize_min_max(data)

    else:
        num_clipped = int(np.sum(data > +1.0)) + int(np.sum(data < -1.0))
        if num_clipped > 0:
            msg = f"Audio data contains {num_clipped} clipped samples (out of {len(data)})"
            if clip == "error":
                raise ValueError(msg)
            elif clip == "warn":
                print(f"WARNING: {msg}")

    Path(path).parent.mkdir(exist_ok=True, parents=True)
    soundfile.write(str(path), data, sr)



from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from pzen.audiofile_utils import soundfile_write
from pzen.signal_utils import (
    Samples,
    Seconds,
    Signal,
    SignalGenerator,
    normalize,
    normalize_min_max,
    pad_to_multiple_of,
)

# -----------------------------------------------------------------------------
# Low-level API
# -----------------------------------------------------------------------------


def test_normalize():
    npt.assert_allclose(
        normalize(np.array([0.5, 0.0, -0.25])),
        np.array([1.0, 0.0, -0.5]),
    )
    npt.assert_allclose(
        normalize(np.array([0.5, 0.5, 0.5])),
        np.array([1.0, 1.0, 1.0]),
    )
    npt.assert_allclose(
        normalize(np.array([0.0, 0.0, 0.0])),
        np.array([0.0, 0.0, 0.0]),
    )


def test_normalize_min_max():
    npt.assert_allclose(
        normalize_min_max(np.array([0.5, 0.0, -0.25])),
        np.array([1.0, -1 / 3, -1.0]),
    )
    npt.assert_allclose(
        normalize_min_max(np.array([0.5, 0.5, 0.5])),
        np.array([0.5, 0.5, 0.5]),
    )
    npt.assert_allclose(
        normalize_min_max(np.array([0.0, 0.0, 0.0])),
        np.array([0.0, 0.0, 0.0]),
    )

    npt.assert_allclose(
        normalize_min_max(np.array([0.0, 0.2, 0.4]), 5, 10),
        np.array([5.0, 7.5, 10.0]),
    )


def test_pad_to_blocksize():
    x = pad_to_multiple_of(np.array([]), 3)
    npt.assert_allclose(x, np.array([]))

    x = pad_to_multiple_of(np.array([1.0]), 3)
    npt.assert_allclose(x, np.array([1.0, 0.0, 0.0]))

    x = pad_to_multiple_of(np.array([1.0, 2.0]), 3)
    npt.assert_allclose(x, np.array([1.0, 2.0, 0.0]))

    x = pad_to_multiple_of(np.array([1.0, 2.0, 3.0]), 3)
    npt.assert_allclose(x, np.array([1.0, 2.0, 3.0]))


# -----------------------------------------------------------------------------
# High-level API
# -----------------------------------------------------------------------------


@pytest.fixture(params=[22050, 44100])
def sr(request: pytest.FixtureRequest) -> int:
    sr_value = request.param
    assert isinstance(sr_value, int)
    return sr_value


def test_signal__add():
    a = Signal(x=np.array([1.0, 0.0, 0.0]), sr=1)
    b = Signal(x=np.array([0.0, 1.0, 0.0]), sr=1)
    npt.assert_allclose((a + b).x, [1.0, 1.0, 0.0])


def test_signal__pad(sr: int):
    x = Signal.empty(sr)
    assert len(x) == 0
    assert x.pad().len() == 0
    assert x.pad(pad_l=Samples(10)).len() == 10
    assert x.pad(pad_r=Samples(20)).len() == 20
    assert x.pad(pad_l=Samples(10), pad_r=Samples(20)).len() == 30

    x = Signal.zeros(Seconds(1.0), sr)
    assert len(x) == sr
    assert x.pad().len() == sr
    assert x.pad(pad_l=Samples(10)).len() == sr + 10
    assert x.pad(pad_r=Samples(20)).len() == sr + 20
    assert x.pad(pad_l=Samples(10), pad_r=Samples(20)).len() == sr + 30


def test_signal__mix_at():
    a = Signal(x=np.array([1.0, 2.0, 3.0]), sr=1)
    b = Signal(x=np.array([1.0]), sr=1)

    npt.assert_allclose(a.mix_at(Samples(0), b).x, [2.0, 2.0, 3.0])
    npt.assert_allclose(a.mix_at(Samples(1), b).x, [1.0, 3.0, 3.0])
    npt.assert_allclose(a.mix_at(Samples(2), b).x, [1.0, 2.0, 4.0])
    npt.assert_allclose(a.mix_at(Samples(3), b).x, [1.0, 2.0, 3.0, 1.0])
    npt.assert_allclose(a.mix_at(Samples(4), b).x, [1.0, 2.0, 3.0, 0.0, 1.0])

    npt.assert_allclose(a.x, [1.0, 2.0, 3.0])

    a.mix_at(Samples(2), b, allow_extend=False)
    with pytest.raises(ValueError):
        a.mix_at(Samples(3), b, allow_extend=False)

    b = Signal(x=np.array([1.0, 1.0]), sr=1)

    npt.assert_allclose(a.mix_at(Samples(0), b).x, [2.0, 3.0, 3.0])
    npt.assert_allclose(a.mix_at(Samples(1), b).x, [1.0, 3.0, 4.0])
    npt.assert_allclose(a.mix_at(Samples(2), b).x, [1.0, 2.0, 4.0, 1.0])
    npt.assert_allclose(a.mix_at(Samples(3), b).x, [1.0, 2.0, 3.0, 1.0, 1.0])
    npt.assert_allclose(a.mix_at(Samples(4), b).x, [1.0, 2.0, 3.0, 0.0, 1.0, 1.0])


# -----------------------------------------------------------------------------
# Generator API
# -----------------------------------------------------------------------------


def test_signal_generator__consistent_dtypes():
    gen = SignalGenerator(sr=22050)
    dtypes = [
        gen.empty().x.dtype,
        gen.silence().x.dtype,
        gen.sine().x.dtype,
    ]
    assert len(set(dtypes)) == 1


def test_signal_generator__envelope_ramped():
    gen = SignalGenerator(sr=22050)
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(0)).x,
        [1.0] * 8,
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(1)).x,
        [1.0] * 8,
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(2)).x,
        [0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.5],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(3)).x,
        [1 / 3, 2 / 3, 1.0, 1.0, 1.0, 1.0, 2 / 3, 1 / 3],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(4)).x,
        [0.25, 0.5, 0.75, 1.0, 1.0, 0.75, 0.5, 0.25],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(5)).x,
        [0.2, 0.4, 0.6, 0.8, 0.8, 0.6, 0.4, 0.2],
    )

    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(3), Samples(4)).x,
        [1 / 3, 2 / 3, 1.0, 1.0, 1.0, 0.75, 0.5, 0.25],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(7), Samples(3), Samples(4)).x,
        [1 / 3, 2 / 3, 1.0, 1.0, 0.75, 0.5, 0.25],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(6), Samples(3), Samples(4)).x,
        [1 / 3, 2 / 3, 1.0, 0.75, 0.5, 0.25],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(5), Samples(3), Samples(4)).x,
        [1 / 3, 2 / 3, 0.75, 0.5, 0.25],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(4), Samples(3), Samples(4)).x,
        [1 / 3, 2 / 3, 0.5, 0.25],
    )

    npt.assert_allclose(
        gen.envelope_ramped(Samples(1), Samples(5), Samples(10)).x,
        [0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(2), Samples(5), Samples(10)).x,
        [0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(3), Samples(5), Samples(10)).x,
        [0.2, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(4), Samples(5), Samples(10)).x,
        [0.2, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(5), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(6), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.4, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(7), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.5, 0.4, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(8), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(9), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.6, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],
    )
    npt.assert_allclose(
        gen.envelope_ramped(Samples(10), Samples(5), Samples(10)).x,
        [0.2, 0.4, 0.6, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],
    )


def test_signal_generator__audio_examples():
    out_dir = Path("/tmp/signal_generator")
    gen = SignalGenerator(sr=22050)

    soundfile_write(
        out_dir / "sine_enveloped_lin.wav",
        gen.sine(440).scale(0.5).envelope_ramped("lin", Seconds(0.2)),
    )
    soundfile_write(
        out_dir / "sine_enveloped_exp.wav",
        gen.sine(440).scale(0.5).envelope_ramped("exp", Seconds(0.2)),
    )

    t = Seconds(10)
    soundfile_write(
        out_dir / "sine_ramped_lin.wav",
        gen.sine(440, t=t) * gen.ramp(t, 0.0, 1.0),
    )
    soundfile_write(
        out_dir / "sine_ramped_exp.wav",
        gen.sine(440, t=t) * gen.ramp(t, 0.0, 1.0).into_exp_envelope(),
    )

    t = Seconds(1)
    soundfile_write(
        out_dir / "sine_from_freqs.wav",
        gen.sine_from_freqs(
            gen.full(t, 220.0).concat(gen.full(t, 440.0)).concat(gen.full(t, 880.0))
        ).scale(0.5),
    )

    t = Seconds(3)
    soundfile_write(
        out_dir / "vibrato_1.wav",
        gen.sine_from_freqs(gen.vibrato(t, 440.0)).scale(0.5),
    )
    soundfile_write(
        out_dir / "vibrato_2.wav",
        gen.sine_from_freqs(gen.vibrato(t, 440.0, f_vibrato=8)).scale(0.5),
    )
    soundfile_write(
        out_dir / "vibrato_3.wav",
        gen.sine_from_freqs(gen.vibrato(t, 440.0, semitones=1.0, t_l=Seconds(1.0))).scale(0.5),
    )

    t = Seconds(5)
    soundfile_write(
        out_dir / "sweep_1.wav",
        gen.sine_from_freqs(gen.sweep(t, 220.0, 880.0)).scale(0.5),
    )
    soundfile_write(
        out_dir / "sweep_2.wav",
        gen.sine_from_freqs(gen.sweep(t, 220.0, 55.0)).scale(0.5),
    )
    soundfile_write(
        out_dir / "sweep_3.wav",
        gen.sine_from_freqs(gen.sweep(t, 20.0, 10000.0)).scale(0.5),
    )
